package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** juxtaposition; zero or more adjacent Elements; can specify a rewriting */
public abstract class Sequence extends Element implements Iterable<Element> {

    // Static Constructors //////////////////////////////////////////////////////////////////////////////

    /** the empty sequence (matches the empty string) */
    public static final Sequence empty = new Sequence.Constant.Empty();

    /** after matching the sequence, do not add anything to the output tree */
    public static Sequence drop(Element[] e, HashSet<Sequence> and, HashSet<Sequence> not, boolean lame) { return new Constant.Drop(e, and, not, lame); }

    /** after matching the sequence, insert a constant into the output tree */
    public static Sequence constant(Element[] e, Object o, HashSet<Sequence> and, HashSet<Sequence> not) { return new Constant(e, o, and, not); }

    /** after matching the sequence, place the result of the <tt>idx</tt>th match in the output tree */
    public static Sequence singleton(Element[] e, int idx, HashSet<Sequence> and, HashSet<Sequence> not) { return new Singleton(e, idx, and, not); }

    /**
     *  after matching the sequence, create the specified output tree
     *  @param tag   the tag for the output tree
     *  @param e     the elements to match
     *  @param drops only elements of <tt>e</tt> whose corresponding <tt>boolean</tt> in <tt>drops</tt> is <i>false</i> will be included in the output tree
     **/
    public static Sequence rewritingSequence(Object tag, Element[] e, boolean[] drops, HashSet<Sequence> and, HashSet<Sequence> not) {
        return new RewritingSequence(tag, e, drops, and, not); }

    ////////////////////////////////////////////////////////////////////////////////

    public Element noFollow = null;
    public final Topology noFollow() { return noFollow==null ? null : noFollow.toAtom(); }

    Topology toAtom() {
        if (elements.length!=1) throw new RuntimeException("cannot invoke toAtom() on a Sequence with " + elements.length + " elements: " + this);
        return elements[0].toAtom();
    }

    protected final Element[] elements;

    final HashSet<Sequence> needs;
    final HashSet<Sequence> hates;
          boolean           lame  = false;

    final Position          firstp;
    Position firstp() { return firstp; }

    public Iterator<Element> iterator()    { return new ArrayIterator<Element>(elements); }
    protected Sequence(Element[] elements, HashSet<Sequence> and, HashSet<Sequence> not) {
        this.needs = and==null ? new HashSet<Sequence>() : and;
        this.hates = not==null ? new HashSet<Sequence>() : not;
        //for(Sequence s : needs) s.lame = true;
        //for(Sequence s : hates) s.lame = true;
        this.elements = elements;
        this.firstp = new Position(0);
    }

    void reachable(HashSet<Position> h) { firstp().reachable(h); }

    Forest epsilonForm() { return firstp().rewrite(null); }

    protected abstract <T> Forest<T> postReduce(Token.Location loc, Forest<T>[] args);


    // Position //////////////////////////////////////////////////////////////////////////////

    /** the imaginary position before or after an element of a sequence; corresponds to an "LR item" */
    public class Position {

        void reachable(HashSet<Position> h) {
            if (h.contains(this)) return;
            h.add(this);
            if (element() != null) element().reachable(h);
        }

                final int      pos;
        private final Position next;
                final Forest[] holder;
        
        private Position(int pos) {
            this.pos      = pos;
            this.next     = pos==elements.length ? null : new Position(pos+1);
            this.holder   = new Forest[elements.length];
        }

        boolean isFirst() { return pos==0; }
        boolean isRightNullable(Walk.Cache cache) {
            if (isLast()) return true;
            if (!element().possiblyEpsilon(cache)) return false;
            return next().isRightNullable(cache);
        }

        /** the element immediately after this Position, or null if this is the last Position */
        public Element  element() { return pos>=elements.length ? null : elements[pos]; }

        /** the element which produces the sequence to which this Position belongs */
        public Sequence owner() { return Sequence.this; }

        /** the next Position (the Position after <tt>this.element()</tt>) */
        public Position next() { return next; }

        /** true iff this Position is the last one in the sequence */
        public boolean isLast() { return next()==null; }

        // Reduction /////////////////////////////////////////////////////////////////////////////////

        <T> Forest<T> rewrite(Token.Location loc) {
            for(int i=pos; i<elements.length; i++) if (holder[i]==null) holder[i] = elements[i].epsilonForm();
            Forest<T> ret = Sequence.this.postReduce(loc, holder);
            for(int k=0; k<pos; k++) holder[k] = null; // to help GC
            return ret;
        }

        public String   toString() {
            StringBuffer ret = new StringBuffer();
            ret.append("<{");
            for(Position p = Sequence.this.firstp(); p != null; p = p.next()) {
                ret.append(' ');
                if (p==this) ret.append(" | ");
                if (p.element()!=null) ret.append(p.element().possiblyEpsilon(null) ? "["+p.element()+"]" : p.element());
                else                   ret.append(' ');
            }
            ret.append("}>");
            return ret.toString();
        }
    }


    // toString //////////////////////////////////////////////////////////////////////////////

    public String toString() { return toString(new StringBuffer(), false).toString(); }
    StringBuffer toString(StringBuffer sb) { return toString(sb, true); }
    StringBuffer toString(StringBuffer sb, boolean spacing) {
        for(int i=0; i<elements.length; i++) {
            sb.append(elements[i].toString());
            sb.append(' ');
        }
        return sb;
    }


    // Specialized Subclasses //////////////////////////////////////////////////////////////////////////////

    static class Constant extends Sequence {
        private final Object result;
        public Constant(Element[] e, Object result, HashSet<Sequence> and, HashSet<Sequence> not) { super(e, and, not); this.result = result; }
        public <T> Forest<T> postReduce(Token.Location loc, Forest<T>[] args) {
            return (Forest<T>)Forest.leaf(loc, result, this);
        }
        static class Drop extends Constant {
            public Drop(Element[] e, HashSet<Sequence> and, HashSet<Sequence> not, boolean lame) {
                super(e, null, and, not);
                this.lame = lame;
            }
        }
        static class Empty extends Sequence.Constant.Drop { public Empty() { super(new Element[] { }, null, null, false); } }
    }

    static class Singleton extends Sequence {
        private final int idx;
        public Singleton(Element e, HashSet<Sequence> and, HashSet<Sequence> not) { this(new Element[] { e }, 0, and, not); }
        public Singleton(Element[] e, int idx, HashSet<Sequence> and, HashSet<Sequence> not) { super(e, and, not); this.idx = idx; }
        public <T> Forest<T> postReduce(Token.Location loc, Forest<T>[] args) { return (Forest<T>)Forest.singleton(loc, args[idx], this); }
    }

    static class Unwrap extends Sequence {
        private boolean[] drops;
        public Unwrap(Element[] e, HashSet<Sequence> and, HashSet<Sequence> not)                  { super(e, and, not); this.drops = null; }
        public Unwrap(Element[] e, boolean[] drops, HashSet<Sequence> and, HashSet<Sequence> not) { super(e, and, not); this.drops = drops; }
        public <T> Forest<T> postReduce(Token.Location loc, Forest<T>[] args) {
            if (drops==null) return Forest.create(loc, null, args, this, true, false);
            int count = 0;
            for(int i=0; i<drops.length; i++) if (!drops[i]) count++;
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, null, args2, this, true, false);            
        }
    }

    static class RewritingSequence extends Sequence {
        private final Object tag;
        private final boolean[] drops;
        private int count = 0;
        public RewritingSequence(Object tag, Element[] e, HashSet<Sequence> and, HashSet<Sequence> not) { this(tag, e, null, and, not); }
        public RewritingSequence(Object tag, Element[] e, boolean[] drops, HashSet<Sequence> and, HashSet<Sequence> not) {
            super(e, and, not);
            this.tag = tag;
            this.drops = drops == null ? new boolean[e.length] : drops;
            for(int i=0; i<this.drops.length; i++) if (!this.drops[i]) count++;
        }
        public <T> Forest<T> postReduce(Token.Location loc, Forest<T>[] args) {
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, (T)tag, args2, this, false, false);
        }
        public StringBuffer toString(StringBuffer sb, boolean spacing) {
            int len = sb.length();
            super.toString(sb, spacing);
            len = sb.length()-len;
            if (spacing) for(int i=0; i<50-len; i++) sb.append(' ');
            sb.append(" => ");
            sb.append(tag);
            return sb;
        }
    }
}
